# Ensure all types are as as statically expected.
import numpy as np

from numba_celltree.constants import (
    FILL_VALUE,
    MAX_N_VERTEX,
    FloatArray,
    FloatDType,
    IntArray,
    IntDType,
)


def cast_vertices(vertices: FloatArray, copy: bool = False) -> FloatArray:
    if isinstance(vertices, np.ndarray):
        vertices = vertices.astype(FloatDType, copy=copy)
    else:
        vertices = np.ascontiguousarray(vertices, dtype=FloatDType)
    if vertices.ndim != 2 or vertices.shape[1] != 2:
        raise ValueError("vertices must have shape (n_points, 2)")
    return vertices


def cast_faces(faces: IntArray, fill_value: int) -> IntArray:
    if isinstance(faces, np.ndarray):
        faces = faces.astype(IntDType, copy=True)
    else:
        faces = np.ascontiguousarray(faces, dtype=IntDType)
    if faces.ndim != 2:
        raise ValueError("faces must have shape (n_face, n_max_vert)")
    _, n_max_vert = faces.shape
    if n_max_vert > MAX_N_VERTEX:
        raise ValueError(
            f"faces contains up to {n_max_vert} vertices for a single face. "
            f"numba_celltree supports a maximum of {MAX_N_VERTEX} vertices. "
            f"Increase MAX_N_VERTEX in the source code, or alter the mesh."
        )
    if fill_value != FILL_VALUE:
        faces[faces == fill_value] = FILL_VALUE
    return faces


def cast_bboxes(bbox_coords: FloatArray) -> FloatArray:
    bbox_coords = np.ascontiguousarray(bbox_coords, dtype=FloatDType)
    if bbox_coords.ndim != 2 or bbox_coords.shape[1] != 4:
        raise ValueError("bbox_coords must have shape (n_box, 4)")
    return bbox_coords


def cast_edges(edges: FloatArray) -> FloatArray:
    edges = np.ascontiguousarray(edges, dtype=FloatDType)
    if edges.ndim != 3 or edges.shape[1] != 2 or edges.shape[2] != 2:
        raise ValueError("edges must have shape (n_edge, 2, 2)")
    return edges



import numba as nb
import numpy as np

from numba_celltree import utils as ut
from numba_celltree.constants import INITIAL_STACK_LENGTH, MAX_N_VERTEX


def test_grow():
    a = np.arange(10)
    b = ut.grow(a)
    assert b.shape == (20,)
    assert np.array_equal(a, b[:10])

    a = np.arange(20).reshape((10, 2))
    b = ut.grow(a)
    assert b.shape == (20, 2)
    assert np.array_equal(a, b[:10])


def test_pop():
    size = 3
    stack = np.arange(10, 10 + size)
    v, size = ut.pop(stack, size)
    assert v == 12
    assert size == 2
    v, size = ut.pop(stack, size)
    assert v == 11
    assert size == 1
    v, size = ut.pop(stack, size)
    assert v == 10
    assert size == 0


def test_push():
    size = 0
    stack = np.empty(3)
    stack, size = ut.push(stack, 0, size)
    stack, size = ut.push(stack, 1, size)
    stack, size = ut.push(stack, 2, size)
    assert size == 3
    assert np.array_equal(stack, [0, 1, 2])


def test_push_size_increases():
    size = 0
    stack = np.empty(1)
    size_before = stack.size
    stack, size = ut.push(stack, 0, size)
    stack, size = ut.push(stack, 1, size)
    assert stack.size == 2
    assert size_before < stack.size
    stack, size = ut.push(stack, 10, size)
    assert stack.size == 4
    assert np.array_equal(stack[:3], [0, 1, 10])


def test_copy():
    dst = np.zeros(5)
    src = np.arange(5)
    ut.copy(src, dst, 3)
    assert np.array_equal(dst, [0, 1, 2, 0, 0])


def test_double_stack():
    stack = ut.allocate_double_stack()
    assert stack.shape == (INITIAL_STACK_LENGTH, 2)

    size = 1
    stack, size = ut.push_both(stack, 1, 2, size)
    assert size == 2
    assert len(stack) == INITIAL_STACK_LENGTH
    assert np.array_equal(stack[1], (1, 2))

    size = INITIAL_STACK_LENGTH
    stack, size = ut.push_both(stack, 1, 2, size)
    assert len(stack) == INITIAL_STACK_LENGTH * 2
    assert size == INITIAL_STACK_LENGTH + 1
    assert np.array_equal(stack[INITIAL_STACK_LENGTH], (1, 2))

    a, b, size = ut.pop_both(stack, size)
    assert size == INITIAL_STACK_LENGTH
    assert len(stack) == INITIAL_STACK_LENGTH * 2
    assert a == 1
    assert b == 2


def test_triple_stack():
    stack = ut.allocate_triple_stack()
    assert stack.shape == (INITIAL_STACK_LENGTH, 3)

    size = 1
    stack, size = ut.push_triple(stack, 1, 2, 3, size)
    assert size == 2
    assert len(stack) == INITIAL_STACK_LENGTH
    assert np.array_equal(stack[1], (1, 2, 3))

    size = INITIAL_STACK_LENGTH
    stack, size = ut.push_triple(stack, 1, 2, 3, size)
    assert len(stack) == INITIAL_STACK_LENGTH * 2
    assert size == INITIAL_STACK_LENGTH + 1
    assert np.array_equal(stack[INITIAL_STACK_LENGTH], (1, 2, 3))

    a, b, c, size = ut.pop_triple(stack, size)
    assert size == INITIAL_STACK_LENGTH
    assert len(stack) == INITIAL_STACK_LENGTH * 2
    assert a == 1
    assert b == 2
    assert c == 3


# These array is not returned properly to dynamic python. This is OK: these
# arrays are exclusively for internal use to temporarily store values.
@nb.njit
def do_allocate_stack():
    stack = ut.allocate_stack()
    return (stack.size == INITIAL_STACK_LENGTH) and (stack[:5].size == 5)


def test_allocate_stack():
    assert do_allocate_stack()


@nb.njit
def do_allocate_polygon():
    poly = ut.allocate_polygon()
    return poly.shape == (MAX_N_VERTEX, 2) and (poly[:5].size == 10)


def test_allocate_polygon():
    assert do_allocate_polygon()


@nb.njit
def do_allocate_clipper():
    clipper = ut.allocate_clip_polygon()
    return clipper.shape == (MAX_N_VERTEX * 2, 2) and (clipper[:5].size == 10)


def test_allocate_clipper():
    assert do_allocate_clipper()



import numpy as np
import pytest

from numba_celltree.algorithms import (
    cohen_sutherland_line_box_clip,
    cyrus_beck_line_polygon_clip,
    liang_barsky_line_box_clip,
)
from numba_celltree.constants import Box, Point


def ab(a, b, c):
    """Flip the result around to compare (a, b) with (b, a)"""
    return (a, c, b)


TOLERANCE = 1e-9


def wraptol(function):
    def f(a, b, poly):
        return function(a, b, poly, TOLERANCE)

    return f


BOX = Box(0.0, 2.0, 0.0, 2.0)
POLY = np.array(
    [
        [0.0, 0.0],
        [2.0, 0.0],
        [2.0, 2.0],
        [0.0, 2.0],
    ]
)
POLY_REVERSED = POLY[::-1, :]


@pytest.mark.parametrize(
    "line_clip, box",
    [
        (cohen_sutherland_line_box_clip, BOX),
        (liang_barsky_line_box_clip, BOX),
        (wraptol(cyrus_beck_line_polygon_clip), POLY),
    ],
)
def test_line_box_clip(line_clip, box):
    a = Point(-1.0, 0.0)
    b = Point(2.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 1.0])
    assert np.allclose(d, [1.0, 2.0])

    a = Point(0.0, -0.1)
    b = Point(0.0, -0.1)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects
    assert np.isnan(c).all()
    assert np.isnan(d).all()

    a = Point(-1.0, 1.0)
    b = Point(3.0, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 1.0])
    assert np.allclose(d, [2.0, 1.0])

    a = Point(1.0, -3.0)
    b = Point(1.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [1.0, 0.0])
    assert np.allclose(d, [1.0, 2.0])

    b = Point(1.0, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [1.0, 0.0])
    assert np.allclose(d, [1.0, 1.0])

    a = Point(1.0, 1.0)
    b = Point(1.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [1.0, 1.0])
    assert np.allclose(d, [1.0, 2.0])

    a = Point(-1.0, 3.0)
    b = Point(3.0, 3.0)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects

    a = Point(-1.0, 1.0)
    b = Point(1.0, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, [0.0, 1.0])
    assert np.allclose(d, [1.0, 1.0])

    # both inside
    a = Point(0.5, 0.5)
    b = Point(1.5, 1.5)
    intersects, c, d = line_clip(a, b, box)
    assert intersects
    assert np.allclose(c, a)
    assert np.allclose(d, b)

    # No intersection, left
    a = Point(-1.5, 0.0)
    b = Point(-0.5, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects

    # No intersection, right
    a = Point(2.5, 0.0)
    b = Point(3.5, 1.0)
    intersects, c, d = line_clip(a, b, box)
    assert not intersects


@pytest.mark.parametrize(
    "line_clip, box",
    [
        (cohen_sutherland_line_box_clip, BOX),
        (liang_barsky_line_box_clip, BOX),
        (wraptol(cyrus_beck_line_polygon_clip), POLY),
        (wraptol(cyrus_beck_line_polygon_clip), POLY_REVERSED),
    ],
)
def test_line_box_clip_degeneracy(line_clip, box):
    def assert_expected(
        a: tuple,
        b: tuple,
        intersects: bool,
        c: tuple = (np.nan, np.nan),
        d: tuple = (np.nan, np.nan),
    ) -> None:
        a = Point(*a)
        b = Point(*b)
        # c, d are the clipped points
        actual, actual_c, actual_d = line_clip(a, b, box)
        print(actual_c, c)
        print(actual_d, d)
        assert intersects is actual
        assert np.allclose(actual_c, c, equal_nan=True)
        assert np.allclose(actual_d, d, equal_nan=True)

        # Direction of the point doesn't change, so neither should the answer.
        actual, actual_c, actual_d = line_clip(a, b, box)
        assert intersects is actual
        assert np.allclose(actual_c, c, equal_nan=True)
        assert np.allclose(actual_d, d, equal_nan=True)

    # Line through vertices
    assert_expected(
        (-1.0, -1.0),
        (3.0, 3.0),
        True,
        (0.0, 0.0),
        (2.0, 2.0),
    )

    # Identity
    assert_expected(
        (-1.0, -1.0),
        (-1.0, -1.0),
        False,
    )

    # Line through lower edge
    assert_expected(
        (-1.0, 0.0),
        (3.0, 0.0),
        True,
        (0.0, 0.0),
        (2.0, 0.0),
    )

    # Line through upper edge
    assert_expected(
        (-1.0, 2.0),
        (3.0, 2.0),
        True,
        (0.0, 2.0),
        (2.0, 2.0),
    )

    # Partial line lower edge
    assert_expected(
        (-1.0, 0.0),
        (1.0, 0.0),
        True,
        (0.0, 0.0),
        (1.0, 0.0),
    )

    # Partial line upper edge
    assert_expected(
        (-1.0, 2.0),
        (1.0, 2.0),
        True,
        (0.0, 2.0),
        (1.0, 2.0),
    )

    # Within lower edge
    assert_expected(
        (0.5, 0.0),
        (1.5, 0.0),
        True,
        (0.5, 0.0),
        (1.5, 0.0),
    )

    # Within upper edge
    assert_expected(
        (0.5, 2.0),
        (1.5, 2.0),
        True,
        (0.5, 2.0),
        (1.5, 2.0),
    )

    # Identical to lower edge
    assert_expected(
        (0.0, 0.0),
        (2.0, 0.0),
        True,
        (0.0, 0.0),
        (2.0, 0.0),
    )

    # Identical to upper edge
    assert_expected(
        (0.0, 2.0),
        (2.0, 2.0),
        True,
        (0.0, 2.0),
        (2.0, 2.0),
    )

    # Identical to left edge
    assert_expected(
        (0.0, 0.0),
        (0.0, 2.0),
        True,
        (0.0, 0.0),
        (0.0, 2.0),
    )

    # Identical to right edge
    assert_expected(
        (2.0, 0.0),
        (2.0, 2.0),
        True,
        (2.0, 0.0),
        (2.0, 2.0),
    )

    # Within left edge
    assert_expected(
        (0.0, 0.5),
        (0.0, 1.5),
        True,
        (0.0, 0.5),
        (0.0, 1.5),
    )

    # Within right edge
    assert_expected(
        (2.0, 0.5),
        (2.0, 1.5),
        True,
        (2.0, 0.5),
        (2.0, 1.5),
    )

    # Identical to left edge
    assert_expected(
        (0.0, 0.0),
        (0.0, 2.0),
        True,
        (0.0, 0.0),
        (0.0, 2.0),
    )

    # Identical to right edge
    assert_expected(
        (2.0, 0.0),
        (2.0, 2.0),
        True,
        (2.0, 0.0),
        (2.0, 2.0),
    )

    # Diagonal line of length (1, 1), touching upper left corner.
    assert_expected(
        (-1.0, 1.0),
        (0.0, 2.0),
        False,
        (np.nan, np.nan),
        (np.nan, np.nan),
    )



from typing import Tuple

import numba as nb
import numpy as np

from numba_celltree.algorithms import (
    cohen_sutherland_line_box_clip,
    cyrus_beck_line_polygon_clip,
)
from numba_celltree.constants import (
    MIN_TOLERANCE,
    PARALLEL,
    TOLERANCE_FACTOR,
    BoolArray,
    CellTreeData,
    FloatArray,
    FloatDType,
    IntArray,
    IntDType,
)
from numba_celltree.geometry_utils import (
    Box,
    Point,
    as_box,
    as_point,
    box_contained,
    boxes_intersect,
    copy_vertices_into,
    lines_intersect,
    point_in_polygon_or_on_edge,
    point_on_edge,
    to_vector,
)
from numba_celltree.utils import (
    allocate_polygon,
    allocate_stack,
    allocate_triple_stack,
    grow,
    pop,
    pop_triple,
    push,
    push_triple,
)


@nb.njit(inline="always")
def concatenate_indices(
    indices: IntArray, counts: IntArray
) -> Tuple[IntArray, IntArray]:
    total_size = sum(counts)
    ii = np.empty(total_size, dtype=IntDType)
    jj = np.empty(total_size, dtype=IntDType)
    start = 0
    for i, size in enumerate(counts):
        end = start + size
        ii[start:end] = indices[i][:size, 0]
        jj[start:end] = indices[i][:size, 1]
        start = end
    return ii, jj


# Inlining saves about 15% runtime
@nb.njit(inline="always")
def locate_point(point: Point, tree: CellTreeData, tolerance: float):
    stack = allocate_stack()
    polygon_work_array = allocate_polygon()
    stack[0] = 0
    return_value = -1
    size = 1

    point_left = Point(point[0] - tolerance, point[1] - tolerance)
    point_right = Point(point[0] + tolerance, point[1] + tolerance)

    while size > 0:
        node_index, size = pop(stack, size)
        node = tree.nodes[node_index]

        # Check if it's a leaf
        if node["child"] == -1:
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                face = tree.elements[bbox_index]
                # Make sure polygons to test is contiguous (stack allocated) array
                # This saves about 40-50% runtime
                poly = copy_vertices_into(tree.vertices, face, polygon_work_array)
                if point_in_polygon_or_on_edge(point, poly, tolerance):
                    return bbox_index
            continue

        dim = 1 if node["dim"] else 0
        left = point_left[dim] <= node["Lmax"]
        right = point_right[dim] >= node["Rmin"]
        left_child = node["child"]
        right_child = left_child + 1

        if left and right:
            # This heuristic is worthwhile because a point will fall into a
            # single face -- if found, we can stop.
            if (node["Lmax"] - point[dim]) < (point_right[dim] - node["Rmin"]):
                stack, size = push(stack, left_child, size)
                stack, size = push(stack, right_child, size)
            else:
                stack, size = push(stack, right_child, size)
                stack, size = push(stack, left_child, size)
        elif left:
            stack, size = push(stack, left_child, size)
        elif right:
            stack, size = push(stack, right_child, size)

    return return_value


@nb.njit(parallel=PARALLEL, cache=True)
def locate_points(points: FloatArray, tree: CellTreeData, tolerance: float):
    n_points = len(points)
    result = np.empty(n_points, dtype=IntDType)
    for i in nb.prange(n_points):  # pylint: disable=not-an-iterable
        point = as_point(points[i])
        result[i] = locate_point(point, tree, tolerance)
    return result


# Inlining saves about 15% runtime
@nb.njit(inline="always")
def locate_point_on_edge(point: Point, tree: CellTreeData, tolerance: float):
    stack = allocate_stack()
    stack[0] = 0
    return_value = -1
    size = 1
    # TODO: Rename allocate_polygon or assess if we can allocate a smaller array.
    edge_work_array = allocate_polygon()

    point_left = Point(point[0] - tolerance, point[1] - tolerance)
    point_right = Point(point[0] + tolerance, point[1] + tolerance)

    while size > 0:
        node_index, size = pop(stack, size)
        node = tree.nodes[node_index]

        # Check if it's a leaf
        if node["child"] == -1:
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                edge = tree.elements[bbox_index]
                segment = copy_vertices_into(tree.vertices, edge, edge_work_array)
                if point_on_edge(point, segment, tolerance):
                    return bbox_index
            continue

        dim = 1 if node["dim"] else 0
        left = point_left[dim] <= node["Lmax"]
        right = point_right[dim] >= node["Rmin"]
        left_child = node["child"]
        right_child = left_child + 1

        if left and right:
            # This heuristic is worthwhile because a point will fall into a
            # single face -- if found, we can stop.
            if (node["Lmax"] - point[dim]) < (point_right[dim] - node["Rmin"]):
                stack, size = push(stack, left_child, size)
                stack, size = push(stack, right_child, size)
            else:
                stack, size = push(stack, right_child, size)
                stack, size = push(stack, left_child, size)
        elif left:
            stack, size = push(stack, left_child, size)
        elif right:
            stack, size = push(stack, right_child, size)

    return return_value


@nb.njit(parallel=PARALLEL, cache=True)
def locate_points_on_edge(points: FloatArray, tree: CellTreeData, tolerance: float):
    n_points = len(points)
    result = np.empty(n_points, dtype=IntDType)
    for i in nb.prange(n_points):  # pylint: disable=not-an-iterable
        point = as_point(points[i])
        result[i] = locate_point_on_edge(point, tree, tolerance)
    return result


@nb.njit(inline="always")
def locate_box(
    box: Box, tree: CellTreeData, indices: IntArray, indices_size: int, index: int
) -> Tuple[int, int]:
    """
    Search the tree for a single axis-aligned box.

    Parameters
    ----------
    box: Box named tuple
    tree: CellTreeData
    indices: IntArray
        Array for results. Contains ``index`` of the box we're searching for
        the first column, and the index of the box in the celltree (if any) in
        the second.
    indices_size: int
        Current number of filled in values in ``indices``.
    index: int
        Current index of the box we're searching.
    """
    tree_bbox = as_box(tree.bbox)
    if not boxes_intersect(box, tree_bbox):
        return 0, indices_size
    stack = allocate_stack()
    stack[0] = 0
    size = 1
    count = 0
    capacity = len(indices)

    while size > 0:
        node_index, size = pop(stack, size)
        node = tree.nodes[node_index]
        # Check if it's a leaf
        if node["child"] == -1:
            # Iterate over the bboxes in the leaf
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                # As a named tuple: saves about 15% runtime
                leaf_box = as_box(tree.bb_coords[bbox_index])
                if boxes_intersect(box, leaf_box):
                    # Exit if we need to re-allocate the array. Exiting instead
                    # of drawing the re-allocation logic in here makes a
                    # significant runtime difference; seems like numba can
                    # optimize this form better.
                    if indices_size >= capacity:
                        return -1, indices_size
                    indices[indices_size, 0] = index
                    indices[indices_size, 1] = bbox_index
                    indices_size += 1
                    count += 1
        else:
            dim = 1 if node["dim"] else 0
            minimum = 2 * dim
            maximum = 2 * dim + 1
            left = box[minimum] <= node["Lmax"]
            right = box[maximum] >= node["Rmin"]
            left_child = node["child"]
            right_child = left_child + 1

            if left and right:
                stack, size = push(stack, left_child, size)
                stack, size = push(stack, right_child, size)
            elif left:
                stack, size = push(stack, left_child, size)
            elif right:
                stack, size = push(stack, right_child, size)

    return count, indices_size


@nb.njit(cache=True)
def locate_boxes_helper(
    box_coords: FloatArray,
    tree: CellTreeData,
    offset: int,
) -> IntArray:
    n_box = len(box_coords)
    # Ensure the initial indices array isn't too small.
    indices = np.empty((max(n_box, 256), 2), dtype=IntDType)
    total_count = 0
    indices_size = 0
    for box_index in range(n_box):
        box = as_box(box_coords[box_index])
        # Re-allocating here is significantly faster than re-allocating inside
        # of ``locate_box``; presumably because that function is kept simpler and
        # numba can optimize better. Unfortunately, that means we have to keep
        # trying until we succeed here; in most cases, success is immediate as
        # the indices array will have enough capacity.
        while True:
            count, indices_size = locate_box(
                box, tree, indices, indices_size, box_index + offset
            )
            if count != -1:
                break
            # Not enough capacity: grow capacity, discard partial work, retry.
            indices_size = total_count
            indices = grow(indices)
        total_count += count
    return indices, total_count


@nb.njit(cache=True, parallel=PARALLEL)
def locate_boxes(box_coords: FloatArray, tree: CellTreeData, n_chunks: int):
    chunks = np.array_split(box_coords, n_chunks)
    offsets = np.zeros(n_chunks, dtype=IntDType)
    for i, chunk in enumerate(chunks[:-1]):
        offsets[i + 1] = offsets[i] + len(chunk)
    # Setup (dummy) typed list for numba to store parallel results.
    indices = [np.empty((0, 2), dtype=IntDType) for _ in range(n_chunks)]
    counts = np.empty(n_chunks, dtype=IntDType)
    for i in nb.prange(n_chunks):
        indices[i], counts[i] = locate_boxes_helper(chunks[i], tree, offsets[i])
    return concatenate_indices(indices, counts)


@nb.njit(inline="always")
def compute_edge_edge_intersect(
    tree: CellTreeData,
    bbox_index: int,
    a: Point,
    b: Point,
    work_array: np.ndarray,
) -> Tuple[bool, Point, Point]:
    tree_edge = tree.elements[bbox_index]
    vertices = tree.vertices[tree_edge]
    p = as_point(vertices[0])
    q = as_point(vertices[1])
    intersects, x, y = lines_intersect(a, b, p, q)
    c = Point(x, y)
    return intersects, c, c


@nb.njit(inline="always")
def compute_edge_face_intersect(
    tree: CellTreeData,
    bbox_index: int,
    a: Point,
    b: Point,
    work_array: np.ndarray,
) -> Tuple[bool, Point, Point]:
    box = as_box(tree.bb_coords[bbox_index])
    intersects, c, d = cohen_sutherland_line_box_clip(a, b, box)
    if intersects:
        polygon = copy_vertices_into(
            tree.vertices, tree.elements[bbox_index], work_array
        )
        tolerance = max(
            MIN_TOLERANCE,
            TOLERANCE_FACTOR * max(box.xmax - box.xmin, box.ymax - box.ymin),
        )
        intersects, c, d = cyrus_beck_line_polygon_clip(a, b, polygon, tolerance)
    return intersects, c, d


def make_locate_edges(intersection_function: nb.types.Callable) -> nb.types.Callable:
    # Inlining this function drives compilation time through the roof. It's
    # probably also a rather bad idea, given its complexity: compared to looking
    # for either boxes or points, checking is more much complicated by involving
    # two intersection algorithms.
    @nb.njit(inline="never")
    def locate_edge(
        a: Point,
        b: Point,
        tree: CellTreeData,
        indices: IntArray,
        intersections: FloatArray,
        indices_size: int,
        index: int,
    ):
        # Check if the entire mesh intersects with the line segment at all
        tree_bbox = as_box(tree.bbox)
        tree_intersects, _, _ = cohen_sutherland_line_box_clip(a, b, tree_bbox)
        if not tree_intersects:
            return 0, indices_size

        V = to_vector(a, b)
        stack = allocate_stack()
        polygon_work_array = allocate_polygon()
        stack[0] = 0
        size = 1
        count = 0
        capacity = len(indices)

        while size > 0:
            node_index, size = pop(stack, size)
            node = tree.nodes[node_index]

            # Check if it's a leaf
            if node["child"] == -1:
                for i in range(node["ptr"], node["ptr"] + node["size"]):
                    bbox_index = tree.bb_indices[i]
                    intersects, c, d = intersection_function(
                        tree, bbox_index, a, b, polygon_work_array
                    )
                    if intersects:
                        # If insufficient capacity, exit.
                        if indices_size >= capacity:
                            return -1, indices_size
                        indices[indices_size, 0] = index
                        indices[indices_size, 1] = bbox_index
                        intersections[indices_size, 0, 0] = c.x
                        intersections[indices_size, 0, 1] = c.y
                        intersections[indices_size, 1, 0] = d.x
                        intersections[indices_size, 1, 1] = d.y
                        indices_size += 1
                        count += 1
                continue

            # Note, "x" is a placeholder for x, y here
            # Contrast with t, which is along vector
            node_dim = 1 if node["dim"] else 0
            dx = V[node_dim]
            if dx > 0.0:
                dx_left = node["Lmax"] - a[node_dim]
                dx_right = node["Rmin"] - b[node_dim]
            else:
                dx_left = node["Lmax"] - b[node_dim]
                dx_right = node["Rmin"] - a[node_dim]

            # Check how origin (a) and end (b) are located compared to box edges
            # (Lmax, Rmin). The box should be investigated if:
            # * the origin is left of Lmax (dx_left >= 0)
            # * the end is right of Rmin (dx_right <= 0)
            left = dx_left >= 0.0
            right = dx_right <= 0.0
            # Now find the intersection coordinates. These have to occur within in
            # the bounds of the vector. Note that if the line has no slope in this
            # dim (dx == 0), we cannot compute the intersection, and we have to
            # defer to the child nodes.
            if dx > 0.0:  # TODO: abs(dx) > EPISLON?
                if left:
                    t_left = dx_left / dx
                    left = t_left >= 0.0
                if right:
                    t_right = dx_right / dx
                    right = t_right <= 1.0
            elif dx < 0.0:
                if left:
                    t_left = 1.0 - (dx_left / dx)
                    left = t_left >= 0.0
                if right:
                    t_right = 1.0 - (dx_right / dx)
                    right = t_right <= 1.0
            # else dx == 0.0. In this case there's no info to extract from this
            # node. We'll fully defer to the children.
            left_child = node["child"]
            right_child = left_child + 1

            if left and right:
                stack, size = push(stack, left_child, size)
                stack, size = push(stack, right_child, size)
            elif left:
                stack, size = push(stack, left_child, size)
            elif right:
                stack, size = push(stack, right_child, size)

        return count, indices_size

    @nb.njit(cache=True)
    def locate_edges_helper(
        edge_coords: FloatArray,
        tree: CellTreeData,
        offset: int,
    ) -> IntArray:
        n_edge = len(edge_coords)
        # Ensure the initial indices array isn't too small.
        n = max(n_edge, 256)
        indices = np.empty((n, 2), dtype=IntDType)
        xy = np.empty((n, 2, 2), dtype=FloatDType)

        total_count = 0
        indices_size = 0
        for edge_index in range(n_edge):
            a = as_point(edge_coords[edge_index, 0])
            b = as_point(edge_coords[edge_index, 1])

            while True:
                count, indices_size = locate_edge(
                    a,
                    b,
                    tree,
                    indices,
                    xy,
                    indices_size,
                    edge_index + offset,
                )
                if count != -1:
                    break
                # Not enough capacity: grow capacity, discard partial work, retry.
                indices_size = total_count
                indices = grow(indices)
                xy = grow(xy)

            total_count += count

        return indices, xy, total_count

    @nb.njit(cache=True, parallel=PARALLEL)
    def locate_edges(box_coords: FloatArray, tree: CellTreeData, n_chunks: int):
        chunks = np.array_split(box_coords, n_chunks)
        offsets = np.zeros(n_chunks, dtype=IntDType)
        for i, chunk in enumerate(chunks[:-1]):
            offsets[i + 1] = offsets[i] + len(chunk)

        # Setup (dummy) typed lists for numba to store parallel results.
        indices = [np.empty((0, 2), dtype=IntDType) for _ in range(n_chunks)]
        intersections = [np.empty((0, 2, 2), dtype=FloatDType) for _ in range(n_chunks)]
        counts = np.empty(n_chunks, dtype=IntDType)
        for i in nb.prange(n_chunks):
            indices[i], intersections[i], counts[i] = locate_edges_helper(
                chunks[i],
                tree,
                offsets[i],
            )

        total_size = sum(counts)
        xy = np.empty((total_size, 2, 2), dtype=FloatDType)
        start = 0
        for i, size in enumerate(counts):
            end = start + size
            xy[start:end] = intersections[i][:size]
            start = end

        ii, jj = concatenate_indices(indices, counts)
        return ii, jj, xy

    return {
        "locate_edges": locate_edges,
        "locate_edges_helper": locate_edges_helper,
        "locate_edge": locate_edge,
    }


# Use the closure to capture the specific intersection function
edge_edge_functions = make_locate_edges(
    intersection_function=compute_edge_edge_intersect
)
edge_face_functions = make_locate_edges(
    intersection_function=compute_edge_face_intersect
)
# Extract the main functions for normal use
locate_edge_edges = edge_edge_functions["locate_edges"]
locate_edge_faces = edge_face_functions["locate_edges"]

# Expose helper functions for testing
locate_edge_edge = edge_edge_functions["locate_edge"]
locate_edge_face = edge_face_functions["locate_edge"]
locate_edge_edge_helper = edge_edge_functions["locate_edges_helper"]
locate_edge_face_helper = edge_face_functions["locate_edges_helper"]


@nb.njit(cache=True)
def collect_node_bounds(tree: CellTreeData) -> FloatArray:
    # Allocate output array.
    # Per row: xmin, xmax, ymin, ymax
    node_bounds = np.empty((len(tree.nodes), 4), dtype=FloatDType)
    # Set bounds of the first node.
    node_bounds[0, 0] = tree.bbox[0]
    node_bounds[0, 1] = tree.bbox[1]
    node_bounds[0, 2] = tree.bbox[2]
    node_bounds[0, 3] = tree.bbox[3]

    # Stack contains: node_index, parent_index, side (right/left)
    ROOT = 0
    RIGHT = 0
    LEFT = 1
    stack = allocate_triple_stack()
    stack[0, :] = (2, ROOT, RIGHT)  # Right child
    stack[1, :] = (1, ROOT, LEFT)  # Left child
    size = 2

    while size > 0:
        # Collect from stacks
        node_index, parent_index, side, size = pop_triple(stack, size)

        parent = tree.nodes[parent_index]
        bbox = node_bounds[parent_index]
        dim = 1 if parent["dim"] else 0

        # Set parent bounding box first.
        # Then place the single new value for the child.
        node_bounds[node_index, 0] = bbox[0]
        node_bounds[node_index, 1] = bbox[1]
        node_bounds[node_index, 2] = bbox[2]
        node_bounds[node_index, 3] = bbox[3]

        if side:
            bound = parent["Lmax"]
        else:
            bound = parent["Rmin"]

        node_bounds[node_index, dim * 2 + side] = bound

        node = tree.nodes[node_index]
        if node["child"] == -1:
            continue

        left_child = node["child"]
        right_child = left_child + 1

        # Right child
        stack, size = push_triple(stack, right_child, node_index, RIGHT, size)
        # Left child
        stack, size = push_triple(stack, left_child, node_index, LEFT, size)

    return node_bounds


@nb.njit(cache=True)
def validate_node_bounds(tree: CellTreeData, node_bounds: FloatArray) -> BoolArray:
    """
    Traverse the tree. Check whether all children are contained in the bounding
    box.

    For the leaf nodes, check whether the bounding boxes are contained.
    """
    node_validity = np.full(len(tree.nodes), False, dtype=np.bool_)
    stack = allocate_stack()
    stack[0] = 0
    size = 1

    while size > 0:
        node_index, size = pop(stack, size)
        bbox = as_box(node_bounds[node_index])
        node = tree.nodes[node_index]

        # Check if it's a leaf:
        if node["child"] == -1:
            valid = True
            for i in range(node["ptr"], node["ptr"] + node["size"]):
                bbox_index = tree.bb_indices[i]
                leaf_box = as_box(tree.bb_coords[bbox_index])
                valid = valid and box_contained(leaf_box, bbox)
            node_validity[node_index] = valid
            continue

        left_child = node["child"]
        right_child = left_child + 1
        left_box = as_box(node_bounds[left_child])
        right_box = as_box(node_bounds[right_child])
        node_validity[node_index] = box_contained(left_box, bbox) and box_contained(
            right_box, bbox
        )

        stack, size = push(stack, right_child, size)
        stack, size = push(stack, left_child, size)

    return node_validity



import numpy as np

from numba_celltree.celltree_base import bbox_distances, bbox_tree, default_tolerance
from numba_celltree.constants import MIN_TOLERANCE, TOLERANCE_FACTOR

bbox_coords = np.array(
    [
        [1.0, 2.0, 1.0, 2.0],
        [4.0, 5.0, 0.0, 1.0],
        [4.0, 5.0, 2.0, 3.0],
        [-1.0, 0.0, 0.0, 4.0],
        [6.0, 8.0, 0.0, 4.0],
        [0.0, 6.0, -1.0, 0.0],
        [0.0, 6.0, 4.0, 5.0],
    ]
)


def test_default_tolerance():
    bb_diagonal = np.array([2.4, 0.5])
    tolerance = default_tolerance(bb_diagonal)
    expected_value = 2.4 * TOLERANCE_FACTOR
    np.testing.assert_allclose(
        tolerance, expected_value, rtol=0, atol=MIN_TOLERANCE / 1e5
    )

    tolerance = default_tolerance(bb_diagonal / 1e4)
    np.testing.assert_allclose(
        tolerance, MIN_TOLERANCE, rtol=0, atol=MIN_TOLERANCE / 1e5
    )


def test_bbox_tree():
    expected_bbox = np.array([-1.0, 8.0, -1.0, 5.0])
    bbox = bbox_tree(bbox_coords)
    np.testing.assert_allclose(bbox, expected_bbox, rtol=0, atol=1e-5)


def test_bbox_distances():
    expected_distances = np.array(
        [
            [1.0, 1.0, 1.41421356],
            [1.0, 1.0, 1.41421356],
            [1.0, 1.0, 1.41421356],
            [1.0, 4.0, 4.12310563],
            [2.0, 4.0, 4.47213595],
            [6.0, 1.0, 6.08276253],
            [6.0, 1.0, 6.08276253],
        ]
    )
    distances = bbox_distances(bbox_coords)
    np.testing.assert_allclose(distances, expected_distances, rtol=0, atol=1e-5)



import numpy as np

from numba_celltree import EdgeCellTree2d
from numba_celltree.constants import CellTreeData

TOLERANCE_ON_EDGE = 1e-9

vertices = np.array([[0.0, 0.0], [1.0, 0.0], [2.0, 0.0], [2.0, 1.0]], dtype=float)
edges = np.array([[0, 1], [1, 2], [2, 3]], dtype=np.int32)


def test_init():
    tree = EdgeCellTree2d(vertices, edges)
    assert tree.vertices.shape == (4, 2)
    assert tree.edges.shape == (3, 2)
    assert tree.n_buckets == 4
    assert tree.cells_per_leaf == 2
    assert tree.nodes.shape == (3,)
    assert tree.bb_indices.shape == (3,)
    assert tree.bb_coords.shape == (3, 4)
    assert tree.bbox.shape == (4,)
    assert isinstance(tree.celltree_data, CellTreeData)

    np.testing.assert_array_equal(tree.bb_indices, np.array([0, 1, 2], dtype=np.int32))

    expected_bb_coords = np.array(
        [
            [0.0, 1.0, 0.0, 0.0],
            [1.0, 2.0, 0.0, 0.0],
            [2.0, 2.0, 0.0, 1.0],
        ],
        dtype=float,
    )
    np.testing.assert_allclose(
        tree.bb_coords, expected_bb_coords, atol=TOLERANCE_ON_EDGE
    )
    np.testing.assert_allclose(
        tree.bbox, np.array([0.0, 2.0, 0.0, 1.0]), atol=TOLERANCE_ON_EDGE
    )


def test_locate_points():
    tree = EdgeCellTree2d(vertices, edges)
    points = np.array([[0.5, 0.0], [1.5, 0.0], [2.0, 0.5]], dtype=float)
    tree_edge_indices = tree.locate_points(points)
    np.testing.assert_array_equal(
        tree_edge_indices, np.array([0, 1, 2], dtype=np.int32)
    )

    points = np.array([[0.5, 0.5], [1.5, 0.5], [2.0, 0.5]], dtype=float)
    tree_edge_indices = tree.locate_points(points)
    np.testing.assert_array_equal(
        tree_edge_indices, np.array([-1, -1, 2], dtype=np.int32)
    )


def test_locate_points_big_coords__tolerance():
    """Small test case extracted from LHM"""
    vertices = np.array(
        [[171805.657000002, 563516.366], [171889.594000001, 563437.333000001]]
    )
    edges = np.array([[0, 1]])
    tree = EdgeCellTree2d(vertices, edges)
    points = np.array([[171882.49385095935, 563444.0183244612]])
    tree_edge_indices = tree.locate_points(points, tolerance=1e-8)
    np.testing.assert_array_equal(tree_edge_indices, np.array([0], dtype=np.int32))
    # test with a tolerance that is too small
    tree_edge_indices = tree.locate_points(points, tolerance=1e-9)
    np.testing.assert_array_equal(tree_edge_indices, np.array([0], dtype=np.int32))


def test_intersect_edges():
    tree = EdgeCellTree2d(vertices, edges)
    edge_coords = np.array(
        [
            [[1.0, -1.0], [1.0, 1.0]],  # 0 orthogonal, on vertex
            [[3.0, 1.0], [-1.0, -1.0]],  # 1 two intersctions
            [[0.0, -1.0], [0.0, 1.0]],  # 2 on start vertex tree
            [[-2.0, -1.0], [-3.0, -1.0]],  # no interesect
        ]
    )
    actual_edge, actual_tree_edge, actual_xy = tree.intersect_edges(edge_coords)
    expected_edge = np.array([0, 0, 1, 1, 2], dtype=np.int32)
    expected_tree_edge = np.array([1, 0, 1, 2, 0], dtype=np.int32)
    expected_xy = np.array(
        [[1.0, 0.0], [1.0, 0.0], [1.0, 0.0], [2.0, 0.5], [0.0, 0.0]], dtype=float
    )

    np.testing.assert_array_equal(actual_edge, expected_edge)
    np.testing.assert_array_equal(actual_tree_edge, expected_tree_edge)
    np.testing.assert_allclose(actual_xy, expected_xy, atol=TOLERANCE_ON_EDGE)



from typing import Sequence, Tuple

import numba as nb
import numpy as np

from numba_celltree.constants import (
    FILL_VALUE,
    MIN_TOLERANCE,
    NDIM,
    PARALLEL,
    TOLERANCE_FACTOR,
    Box,
    FloatArray,
    FloatDType,
    IntArray,
    Point,
    Triangle,
    Vector,
)
from numba_celltree.utils import allocate_box_polygon, allocate_polygon


@nb.njit(inline="always")
def to_vector(a: Point, b: Point) -> Vector:
    return Vector(b.x - a.x, b.y - a.y)


@nb.njit(inline="always")
def as_point(a: FloatArray) -> Point:
    return Point(a[0], a[1])


@nb.njit(inline="always")
def as_box(arr: FloatArray) -> Box:
    return Box(
        arr[0],
        arr[1],
        arr[2],
        arr[3],
    )


@nb.njit(inline="always")
def as_triangle(vertices: FloatArray, face: IntArray) -> Triangle:
    return Triangle(
        as_point(vertices[face[0]]),
        as_point(vertices[face[1]]),
        as_point(vertices[face[2]]),
    )


@nb.njit(inline="always")
def to_point(t: float, a: Point, V: Vector) -> Point:
    return Point(a.x + t * V.x, a.y + t * V.y)


@nb.njit(inline="always")
def cross_product(u: Vector, v: Vector) -> float:
    return u.x * v.y - u.y * v.x


@nb.njit(inline="always")
def dot_product(u: Vector, v: Vector) -> float:
    return u.x * v.x + u.y * v.y


@nb.njit(inline="always")
def polygon_length(face: IntArray) -> int:
    # A minimal polygon is a triangle
    n = len(face)
    for i in range(3, n):
        if face[i] == FILL_VALUE:
            return i
    return n


@nb.njit(inline="always")
def polygon_area(polygon: Sequence) -> float:
    length = len(polygon)
    area = 0.0
    a = Point(polygon[0][0], polygon[0][1])
    b = Point(polygon[1][0], polygon[1][1])
    U = to_vector(a, b)
    for i in range(2, length):
        c = Point(polygon[i][0], polygon[i][1])
        V = to_vector(c, a)
        area += abs(cross_product(U, V))
        b = c
        U = V
    return 0.5 * area


@nb.njit(inline="always")
def point_in_polygon(p: Point, poly: Sequence) -> bool:
    # Refer to: https://wrf.ecse.rpi.edu/Research/Short_Notes/pnpoly.html
    # Copyright (c) 1970-2003, Wm. Randolph Franklin
    # MIT license.
    #
    # Quote:
    # > I run a semi-infinite ray horizontally (increasing x, fixed y) out from
    # > the test point, and count how many edges it crosses. At each crossing,
    # > the ray switches between inside and outside. This is called the Jordan
    # > curve theorem.
    # >
    # > The case of the ray going thru a vertex is handled correctly via a
    # > careful selection of inequalities. Don't mess with this code unless
    # > you're familiar with the idea of Simulation of Simplicity. This pretends
    # > to shift the ray infinitesimally down so that it either clearly
    # > intersects, or clearly doesn't touch. Since this is merely a conceptual,
    # > infinitesimal, shift, it never creates an intersection that didn't exist
    # > before, and never destroys an intersection that clearly existed before.
    # >
    # > The ray is tested against each edge thus:
    # > 1. Is the point in the half-plane to the left of the extended edge? and
    # > 2. Is the point's Y coordinate within the edge's Y-range?
    # >
    # > Handling endpoints here is tricky.
    #
    # For the Simulation of Simplicity concept, see:
    # Edelsbrunner, H., & Mücke, E. P. (1990). Simulation of simplicity: a
    # technique to cope with degenerate cases in geometric algorithms. ACM
    # Transactions on Graphics (tog), 9(1), 66-104.
    #
    # In this case, this guarantees there will be no "on-edge" answers, which
    # are degenerative. For another application of simulation of simplicity,
    # see:
    # Rappoport, A. (1991). An efficient algorithm for line and polygon
    # clipping. The Visual Computer, 7(1), 19-28.
    length = len(poly)
    v0 = as_point(poly[-1])
    c = False
    for i in range(length):
        v1 = as_point(poly[i])
        # Do not split this in two conditionals: if the first conditional fails,
        # the second will not be executed in Python's (and C's) execution model.
        # This matters because the second can result in division by zero.
        if (v0.y > p.y) != (v1.y > p.y) and p.x < (
            (v1.x - v0.x) * (p.y - v0.y) / (v1.y - v0.y) + v0.x
        ):
            c = not c
        v0 = v1
    return c


@nb.njit(inline="always")
def in_bounds(p: Point, a: Point, b: Point, tolerance: float) -> bool:
    """
    Check whether point p falls within the bounding box created by a and b
    (after we've checked the size of the cross product).
    However, we must take into account that a line may be either vertical
    (dx=0) or horizontal (dy=0) and only evaluate the non-zero value.
    If the area created by p, a, b is tiny AND p is within the bounds of a and
    b, the point lies very close to the edge.

    This is a branchless implementation.
    """
    xmin = min(a.x, b.x) - tolerance
    xmax = max(a.x, b.x) + tolerance
    ymin = min(a.y, b.y) - tolerance
    ymax = max(a.y, b.y) + tolerance
    dx = xmax - xmin
    dy = ymax - ymin
    # Determine which bound to use based on which dimension is larger
    use_x_bound = abs(dx) >= abs(dy)
    # Combine results without branching
    return (use_x_bound and ((p.x >= xmin) and (p.x <= xmax))) or (
        not use_x_bound and ((p.y >= ymin) and (p.y <= ymax))
    )


@nb.njit(inline="always")
def point_in_polygon_or_on_edge(p: Point, poly: FloatArray, tolerance: float) -> bool:
    length = len(poly)
    v0 = as_point(poly[-1])
    U = to_vector(p, v0)
    c = False
    for i in range(length):
        v1 = as_point(poly[i])
        if v1 == v0:
            continue
        V = to_vector(p, v1)
        # Compute the (twofold) area of formed by the point (p) and two
        # vertices of the polygon (v0, v1). If this area is extremely small,
        # the point is (nearly) on the edge, or it is collinear. We can test if
        # if's collinear by checking whether it falls in the bounding box of
        # points v0 and v1.
        A = cross_product(U, V)
        W = to_vector(v0, v1)
        L2 = W.x * W.x + W.y * W.y
        # Compute optimized equivalent of A/length < tolerance (no sqrt, no
        # division).
        if (A * A) < ((tolerance * L2) * tolerance) and in_bounds(p, v0, v1, tolerance):
            return True

        if (v0.y > p.y) != (v1.y > p.y) and p.x < (
            (v1.x - v0.x) * (p.y - v0.y) / (v1.y - v0.y) + v0.x
        ):
            c = not c

        v0 = v1
        U = V
    return c


@nb.njit(inline="always")
def point_on_edge(p: Point, edge: FloatArray, tolerance: float) -> bool:
    v0 = as_point(edge[0])
    v1 = as_point(edge[1])
    if v1 == v0:
        return False
    U = to_vector(p, v0)
    V = to_vector(p, v1)
    W = to_vector(v0, v1)
    L2 = W.x * W.x + W.y * W.y
    A = cross_product(U, V)
    # Compute optimized equivalent of A/length < tolerance (no sqrt, no
    # division).
    if (A * A) < ((tolerance * L2) * tolerance) and in_bounds(p, v0, v1, tolerance):
        return True
    return False


@nb.njit(inline="always")
def boxes_intersect(a: Box, b: Box) -> bool:
    """
    Parameters
    ----------
    a: (xmin, xmax, ymin, ymax)
    b: (xmin, xmax, ymin, ymax)
    """
    return a.xmin < b.xmax and b.xmin < a.xmax and a.ymin < b.ymax and b.ymin < a.ymax


@nb.njit(inline="always")
def box_contained(a: Box, b: Box) -> bool:
    """
    Whether a is contained by b.

    Parameters
    ----------
    a: (xmin, xmax, ymin, ymax)
    b: (xmin, xmax, ymin, ymax)
    """
    return (
        a.xmin >= b.xmin and a.xmax <= b.xmax and a.ymin >= b.ymin and a.ymax <= b.ymax
    )


@nb.njit(inline="always")
def left_of(a: Point, p: Point, U: Vector) -> bool:
    # Whether point a is left of vector U
    # U: p -> q direction vector
    return U.x * (a.y - p.y) > U.y * (a.x - p.x)


@nb.njit(inline="always")
def has_overlap(a: float, b: float, p: float, q: float, tolerance: float) -> bool:
    return ((min(a, b) - max(p, q)) < tolerance) and (
        (min(p, q) - max(a, b)) < tolerance
    )


@nb.njit(inline="always")
def intersection_location_point(
    V: Vector, U: Vector, a: Point, p: Point, tolerance: float
) -> Point:
    # Calculate intersection point
    denom = cross_product(V, U)
    if abs(denom) < tolerance:
        return np.nan, np.nan  # Parallel lines

    R = to_vector(a, p)
    t = cross_product(R, U) / denom
    x = a.x + t * V.x
    y = a.y + t * V.y
    return x, y


@nb.njit(inline="always")
def midpoint_collinear_lines(a: Point, b: Point, p: Point, q: Point) -> Point:
    """
    Calculate the midpoint of the overlapping portion of two collinear line segments.
    If the segments do not overlap, return a Point with NaN coordinates.
    """
    # Ensure the points are ordered (smallest to largest) along the x-axis or y-axis
    if a.x > b.x or (a.x == b.x and a.y > b.y):
        a, b = b, a
    if p.x > q.x or (p.x == q.x and p.y > q.y):
        p, q = q, p

    # Find the overlapping segment
    overlap_start_x = max(a.x, p.x)
    overlap_start_y = max(a.y, p.y)
    overlap_end_x = min(b.x, q.x)
    overlap_end_y = min(b.y, q.y)

    # Check if there is an overlap
    if overlap_start_x > overlap_end_x or overlap_start_y > overlap_end_y:
        # No overlap
        return np.nan, np.nan

    # Compute the midpoint of the overlapping segment
    midpoint_x = 0.5 * (overlap_start_x + overlap_end_x)
    midpoint_y = 0.5 * (overlap_start_y + overlap_end_y)
    return midpoint_x, midpoint_y


@nb.njit(inline="always")
def lines_intersect(
    a: Point, b: Point, p: Point, q: Point
) -> tuple[bool, float, float]:
    """Test whether line segment a -> b intersects p -> q."""
    V = to_vector(a, b)
    U = to_vector(p, q)

    # No intersection if no length.
    if (U.x == 0 and U.y == 0) or (V.x == 0 and V.y == 0):
        return False, np.nan, np.nan
    # If x- or y-components are zero, they can only intersect if x or y is identical.
    if (U.x == 0) and (V.x == 0) and a.x != p.x:
        return False, np.nan, np.nan
    if (U.y == 0) and (V.y == 0) and a.y != p.y:
        return False, np.nan, np.nan

    tolerance = max(MIN_TOLERANCE, TOLERANCE_FACTOR * max(abs(U.x), abs(U.y)))
    # bounds check
    if (not has_overlap(a.x, b.x, p.x, q.x, tolerance)) or (
        not has_overlap(a.y, b.y, p.y, q.y, tolerance)
    ):
        return False, np.nan, np.nan

    # Check a and b for separation by U (p -> q)
    # and p and q for separation by V (a -> b)
    if (left_of(a, p, U) != left_of(b, p, U)) and (
        left_of(p, a, V) != left_of(q, a, V)
    ):
        x, y = intersection_location_point(V, U, a, p, tolerance)
        return True, x, y

    # Detect collinear case, where segments lie on the same infite line.
    R = to_vector(a, p)
    S = to_vector(a, q)
    if (abs(cross_product(V, R)) < tolerance) and (
        abs(cross_product(V, S) < tolerance)
    ):
        x, y = midpoint_collinear_lines(a, b, p, q)
        return True, x, y

    return False, np.nan, np.nan


@nb.njit(inline="always")
def bounding_box(
    polygon: IntArray, vertices: FloatArray
) -> Tuple[float, float, float, float]:
    max_n_verts = len(polygon)
    first_vertex = vertices[polygon[0]]
    xmin = xmax = first_vertex[0]
    ymin = ymax = first_vertex[1]
    for i in range(1, max_n_verts):
        index = polygon[i]
        if index == FILL_VALUE:
            break
        vertex = vertices[index]
        x = vertex[0]
        y = vertex[1]
        xmin = min(xmin, x)
        xmax = max(xmax, x)
        ymin = min(ymin, y)
        ymax = max(ymax, y)
    return (xmin, xmax, ymin, ymax)


@nb.njit(cache=True)
def build_face_bboxes(
    faces: IntArray,
    vertices: FloatArray,
) -> FloatArray:
    # Make room for the bounding box of every polygon.
    n_polys = len(faces)
    bbox_coords = np.empty((n_polys, NDIM * 2), FloatDType)

    for i in nb.prange(n_polys):  # pylint: disable=not-an-iterable
        polygon = faces[i]
        bbox_coords[i] = bounding_box(polygon, vertices)

    return bbox_coords


@nb.njit(inline="always")
def edge_bounding_box(
    edge: IntArray, vertices: FloatArray
) -> Tuple[float, float, float, float]:
    x0, y0 = vertices[edge[0]]
    x1, y1 = vertices[edge[1]]
    dx = abs(x1 - x0)
    dy = abs(y1 - y0)
    tolerance = max(MIN_TOLERANCE, TOLERANCE_FACTOR * max(dx, dy))
    # Edges may be axis-aligned. Create a fictitious width in this case.
    xmin = min(x0 - tolerance, x1 - tolerance)
    xmax = max(x0 + tolerance, x1 + tolerance)
    ymin = min(y0 - tolerance, y1 - tolerance)
    ymax = max(y0 + tolerance, y1 + tolerance)
    return (xmin, xmax, ymin, ymax)


@nb.njit(cache=True)
def build_edge_bboxes(
    edges: IntArray,
    vertices: FloatArray,
) -> FloatArray:
    # Make room for the bounding box of every polygon.
    n_polys = len(edges)
    bbox_coords = np.empty((n_polys, NDIM * 2), FloatDType)

    for i in nb.prange(n_polys):  # pylint: disable=not-an-iterable
        edge = edges[i]
        bbox_coords[i] = edge_bounding_box(edge, vertices)

    return bbox_coords


@nb.njit(inline="always")
def copy_vertices(vertices: FloatArray, face: IntArray) -> FloatArray:
    length = polygon_length(face)
    out = allocate_polygon()
    for i in range(length):
        v = vertices[face[i]]
        out[i, 0] = v[0]
        out[i, 1] = v[1]
    return out[:length]


@nb.njit(inline="always")
def copy_vertices_into(
    vertices: FloatArray, face: IntArray, out: FloatArray
) -> FloatArray:
    length = polygon_length(face)
    for i in range(length):
        v = vertices[face[i]]
        out[i, 0] = v[0]
        out[i, 1] = v[1]
    return out[:length]


@nb.njit(inline="always")
def copy_box_vertices(box: Box) -> FloatArray:
    a = allocate_box_polygon()
    a[0, 0] = box.xmin
    a[0, 1] = box.ymin
    a[1, 0] = box.xmax
    a[1, 1] = box.ymin
    a[2, 0] = box.xmax
    a[2, 1] = box.ymax
    a[3, 0] = box.xmin
    a[3, 1] = box.ymax
    return a


@nb.njit(inline="always")
def point_inside_box(a: Point, box: Box):
    return box.xmin < a.x and a.x < box.xmax and box.ymin < a.y and a.y < box.ymax


@nb.njit(inline="always")
def flip(face: IntArray, length: int) -> None:
    end = length - 1
    for i in range(int(length / 2)):
        j = end - i
        face[i], face[j] = face[j], face[i]
    return


@nb.njit(parallel=PARALLEL, cache=True)
def counter_clockwise(vertices: FloatArray, faces: IntArray) -> None:
    n_face = len(faces)
    for i_face in nb.prange(n_face):
        face = faces[i_face]
        length = polygon_length(face)
        a = as_point(vertices[face[length - 2]])
        b = as_point(vertices[face[length - 1]])
        for i in range(length):
            c = as_point(vertices[face[i]])
            u = to_vector(a, b)
            v = to_vector(a, c)
            product = cross_product(u, v)
            if product == 0:
                a = b
                b = c
            elif product < 0:
                flip(face, length)
            else:
                break
    return



import abc
from typing import Optional

import numpy as np

from numba_celltree.constants import (
    MIN_TOLERANCE,
    TOLERANCE_FACTOR,
    BoolArray,
    FloatArray,
    FloatDType,
    IntArray,
)
from numba_celltree.query import (
    collect_node_bounds,
    validate_node_bounds,
)


def bbox_tree(bb_coords: FloatArray) -> FloatArray:
    xmin = bb_coords[:, 0].min()
    xmax = bb_coords[:, 1].max()
    ymin = bb_coords[:, 2].min()
    ymax = bb_coords[:, 3].max()
    return np.array([xmin, xmax, ymin, ymax], dtype=FloatDType)


def bbox_distances(bb_coords: FloatArray) -> FloatArray:
    """Compute the dx, dy and dxy distances for the bounding boxes.

    Parameters
    ----------
    bb_coords: np.ndarray of shape (n_nodes, 4)
        The bounding box coordinates of the nodes.

    Returns
    -------
    distances: np.ndarray of shape (n_nodes, 3)
        Respectively the dx, dy and dxy distances for the bounding boxes.
    """
    distances = np.empty((bb_coords.shape[0], 3), dtype=FloatDType)
    # dx
    distances[:, 0] = bb_coords[:, 1] - bb_coords[:, 0]
    # dy
    distances[:, 1] = bb_coords[:, 3] - bb_coords[:, 2]
    # dxy
    distances[:, 2] = np.sqrt(distances[:, 0] ** 2 + distances[:, 1] ** 2)
    return distances


def default_tolerance(bb_diagonal: FloatArray) -> float:
    return max(MIN_TOLERANCE, TOLERANCE_FACTOR * max(bb_diagonal))


class CellTree2dBase(abc.ABC):
    @abc.abstractmethod
    def locate_points(
        self, points: FloatArray, tolerance: Optional[float] = None
    ) -> IntArray:
        pass

    @property
    def node_bounds(self):
        """Return the bounds (xmin, xmax, ymin, ymax) for every node of the tree."""
        return collect_node_bounds(self.celltree_data)

    def validate_node_bounds(self) -> BoolArray:
        """
        Traverse the tree. Check whether all children are contained in the bounding
        box.

        For the leaf nodes, check whether the bounding boxes are contained.

        Returns
        -------
        node_validity: np.array of bool
            For each node, whether all children are fully contained by its
            bounds.
        """
        return validate_node_bounds(self.celltree_data, self.node_bounds)

    def to_dict_of_lists(self):
        """
        Convert the tree structure to a dict of lists.

        Such a dict can be ingested by e.g. NetworkX to produce visualize the
        tree structure.

        Returns
        -------
        dict_of_lists: Dict[Int, List[Int]]
            Contains for every node a list with its children.

        Examples
        --------
        >>> import networkx
        >>> from networkx.drawing.nx_pydot import graphviz_layout
        >>> d = celltree.to_dict_of_lists()
        >>> G = networkx.DiGraph(d)
        >>> positions = graphviz_layout(G, prog="dot")
        >>> networkx.draw(G, positions, with_labels=True)

        Note that computing the graphviz layout may be quite slow!
        """
        dict_of_lists = {}
        for parent_index, node in enumerate(self.celltree_data.nodes):
            left_child = node["child"]
            if left_child == -1:
                dict_of_lists[parent_index] = []
            else:
                right_child = left_child + 1
                dict_of_lists[parent_index] = [left_child, right_child]

        return dict_of_lists



from typing import Optional, Tuple

import numba as nb
import numpy as np

from numba_celltree.cast import cast_edges, cast_vertices
from numba_celltree.celltree_base import (
    CellTree2dBase,
    bbox_distances,
    bbox_tree,
    default_tolerance,
)
from numba_celltree.constants import (
    CellTreeData,
    FloatArray,
    IntArray,
)
from numba_celltree.creation import initialize
from numba_celltree.geometry_utils import build_edge_bboxes
from numba_celltree.query import locate_edge_edges, locate_points_on_edge


class EdgeCellTree2d(CellTree2dBase):
    """
    Construct a cell tree from 2D vertices and an edges indexing array.

    Parameters
    ----------
    vertices: ndarray of floats with shape ``(n_point, 2)``
        Corner coordinates (x, y) of the cells.
    edges: ndarray of integers with shape ``(n_edge, 2)``
        Index identifying for every edge the indices of its two nodes.
    n_buckets: int, optional, default: 4
        The number of "buckets" used in tree construction. Must be higher
        or equal to 2. Values over 8 provide diminishing returns.
    cells_per_leaf: int, optional, default: 2
        The number of cells in the leaf nodes of the cell tree. Can be set
        to only 1, but this doubles memory footprint for slightly faster
        lookup. Increase this to reduce memory usage at the cost of lookup
        performance.
    """

    def __init__(
        self,
        vertices: FloatArray,
        edges: IntArray,
        n_buckets: int = 4,
        cells_per_leaf: int = 2,
    ):
        # Determine the tolerance for the bounding boxes. This is mainly
        # relevant when edges are axis-aligned.
        if n_buckets < 2:
            raise ValueError("n_buckets must be >= 2")
        if cells_per_leaf < 1:
            raise ValueError("cells_per_leaf must be >= 1")

        vertices = cast_vertices(vertices, copy=True)

        bb_coords = build_edge_bboxes(edges, vertices)
        nodes, bb_indices = initialize(edges, bb_coords, n_buckets, cells_per_leaf)
        self.vertices = vertices
        self.edges = edges
        self.n_buckets = n_buckets
        self.cells_per_leaf = cells_per_leaf
        self.nodes = nodes
        self.bb_indices = bb_indices
        self.bb_coords = bb_coords
        self.bbox = bbox_tree(bb_coords)
        self.bb_distances = bbox_distances(bb_coords)
        self.celltree_data = CellTreeData(
            self.edges,
            self.vertices,
            self.nodes,
            self.bb_indices,
            self.bb_coords,
            self.bbox,
            self.cells_per_leaf,
        )

    def locate_points(
        self, points: FloatArray, tolerance: Optional[float] = None
    ) -> IntArray:
        """
        Find the index of a face that contains a point.

        Parameters
        ----------
        points: ndarray of floats with shape ``(n_point, 2)``
            Coordinates of the points to be located.
        tolerance: float, optional
            The tolerance used to determine whether a point is on an edge. If
            the distance from the point to the edge is smaller than this value,
            the point is considered to be on the edge. If None, the method tries
            to estimate an appropriate tolerance by multiplying the maximum
            diagonal of the bounding boxes with 1e-12.


        Returns
        -------
        tree_edge_indices: ndarray of integers with shape ``(n_point,)``
            For every point, the index of the edge it falls on. Points not
            falling on any edge are marked with a value of ``-1``.
        """
        if tolerance is None:
            tolerance = default_tolerance(self.bb_distances[:, 2])
        points = cast_vertices(points)
        return locate_points_on_edge(points, self.celltree_data, tolerance)

    def intersect_edges(
        self,
        edge_coords: FloatArray,
    ) -> Tuple[IntArray, IntArray, FloatArray]:
        """
        Find the index of an edge intersecting with an edge.

        Parameters
        ----------
        edge_coords: ndarray of floats with shape ``(n_edge, 2, 2)``
            Every row containing ``((x0, y0), (x1, y1))``.

        Returns
        -------
        edge_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the bounding box.
        tree_edge_indices: ndarray of integers with shape ``(n_found,)``
            Indices of the edge.
        intersection: ndarray of floats with shape ``(n_found, 2)``
            Coordinate pair of the intersection.
        """
        edge_coords = cast_edges(edge_coords)
        n_chunks = nb.get_num_threads()
        edge_indices, tree_edge_indices, xy = locate_edge_edges(
            edge_coords, self.celltree_data, n_chunks
        )
        intersection_xy = np.ascontiguousarray(xy[:, 0])
        return edge_indices, tree_edge_indices, intersection_xy
